mod models;
mod repository;
mod schema;

use diesel::connection::InstrumentationEvent;
use diesel::r2d2::CustomizeConnection;
use diesel::Connection;

pub use repository::SqliteRepository;

use diesel::{
    r2d2::{ConnectionManager, Pool},
    SqliteConnection,
};
use diesel_migrations::{embed_migrations, EmbeddedMigrations, MigrationHarness};

use std::{error::Error, fmt, path::Path};

pub const MIGRATIONS: EmbeddedMigrations = embed_migrations!("migrations");

pub type DbPool = Pool<ConnectionManager<SqliteConnection>>;
pub type DbResult<T> = Result<T, DbError>;

#[derive(Debug)]
pub enum DbError {
    Pool(diesel::r2d2::PoolError),
    Db(diesel::result::Error),
}

impl fmt::Display for DbError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Pool(_) => write!(f, "failed to acquire connection from pool"),
            Self::Db(_) => write!(f, "error while executing query"),
        }
    }
}

impl Error for DbError {
    fn source(&self) -> Option<&(dyn Error + 'static)> {
        match self {
            Self::Pool(err) => Some(err),
            Self::Db(err) => Some(err),
        }
    }
}

impl From<diesel::r2d2::PoolError> for DbError {
    fn from(value: diesel::r2d2::PoolError) -> Self {
        Self::Pool(value)
    }
}

impl From<diesel::result::Error> for DbError {
    fn from(value: diesel::result::Error) -> Self {
        Self::Db(value)
    }
}

impl From<DbError> for gpodder::AuthErr {
    fn from(value: DbError) -> Self {
        match value {
            DbError::Pool(err) => Self::Other(Box::new(err)),
            DbError::Db(err) => Self::Other(Box::new(err)),
        }
    }
}

#[derive(Debug)]
pub struct AddQueryDebugLogs;

impl CustomizeConnection<SqliteConnection, diesel::r2d2::Error> for AddQueryDebugLogs {
    fn on_acquire(&self, conn: &mut SqliteConnection) -> Result<(), diesel::r2d2::Error> {
        conn.set_instrumentation(|event: InstrumentationEvent<'_>| match event {
            InstrumentationEvent::StartQuery { query, .. } => {
                tracing::debug!("{}", query);
            }
            _ => {}
        });

        Ok(())
    }
}

pub fn initialize_db(path: impl AsRef<Path>, run_migrations: bool) -> Result<DbPool, DbError> {
    let manager = ConnectionManager::<SqliteConnection>::new(path.as_ref().to_string_lossy());
    let pool = Pool::builder()
        .connection_customizer(Box::new(AddQueryDebugLogs))
        .build(manager)?;

    if run_migrations {
        pool.get()?.run_pending_migrations(MIGRATIONS).unwrap();
    }

    Ok(pool)
}
