use std::time::Duration;

use crate::server;

pub fn serve(config: &crate::config::Config) -> u8 {
    tracing_subscriber::fmt::init();

    tracing::info!("Initializing database and running migrations");

    let store =
        gpodder_sqlite::SqliteRepository::from_path(config.data_dir.join(crate::DB_FILENAME))
            .unwrap();
    let store = gpodder::GpodderRepository::new(store);

    let ctx = server::Context { store };
    let app = server::app(ctx.clone());

    let rt = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .unwrap();

    let address = format!("{}:{}", config.domain, config.port);
    tracing::info!("Starting server on {address}");

    let session_removal_duration = Duration::from_secs(config.session_cleanup_interval);

    rt.block_on(async {
        tokio::task::spawn(async move {
            let mut interval = tokio::time::interval(session_removal_duration);

            loop {
                interval.tick().await;

                tracing::info!("Performing session cleanup");

                match ctx.store.remove_old_sessions() {
                    Ok(n) => {
                        tracing::info!("Removed {} old sessions", n);
                    }
                    Err(err) => {
                        tracing::error!("Error occured during session cleanup: {}", err);
                    }
                }
            }
        });

        let listener = tokio::net::TcpListener::bind(address).await.unwrap();
        axum::serve(listener, app.into_make_service())
            .await
            .unwrap()
    });

    0
}
