use axum::{
    extract::{Path, Query, State},
    middleware,
    routing::post,
    Extension, Json, Router,
};
use chrono::DateTime;
use serde::{Deserialize, Serialize};

use crate::server::{
    error::{AppError, AppResult},
    gpodder::{
        auth_middleware,
        format::{Format, StringWithFormat},
        models,
        models::UpdatedUrlsResponse,
    },
    Context,
};

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .route(
            "/{username}",
            post(post_episode_actions).get(get_episode_actions),
        )
        .layer(middleware::from_fn_with_state(ctx.clone(), auth_middleware))
}

async fn post_episode_actions(
    State(ctx): State<Context>,
    Path(username): Path<StringWithFormat>,
    Extension(user): Extension<gpodder::User>,
    Json(actions): Json<Vec<models::EpisodeAction>>,
) -> AppResult<Json<UpdatedUrlsResponse>> {
    if username.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(tokio::task::spawn_blocking(move || {
        ctx.store
            .add_episode_actions(&user, actions.into_iter().map(Into::into).collect())
    })
    .await
    .unwrap()
    .map(|time_changed| {
        Json(UpdatedUrlsResponse {
            timestamp: time_changed.timestamp(),
            update_urls: Vec::new(),
        })
    })?)
}

#[derive(Deserialize, Default)]
#[serde(default)]
struct FilterQuery {
    podcast: Option<String>,
    device: Option<String>,
    since: Option<i64>,
    aggregated: bool,
}

#[derive(Serialize)]
struct EpisodeActionsResponse {
    timestamp: i64,
    actions: Vec<models::EpisodeAction>,
}

async fn get_episode_actions(
    State(ctx): State<Context>,
    Path(username): Path<StringWithFormat>,
    Extension(user): Extension<gpodder::User>,
    Query(filter): Query<FilterQuery>,
) -> AppResult<Json<EpisodeActionsResponse>> {
    if username.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    let since = filter
        .since
        .map(|ts| DateTime::from_timestamp(ts, 0))
        .flatten();

    Ok(tokio::task::spawn_blocking(move || {
        ctx.store.episode_actions_for_user(
            &user,
            since,
            filter.podcast,
            filter.device,
            filter.aggregated,
        )
    })
    .await
    .unwrap()
    .map(|(ts, actions)| {
        Json(EpisodeActionsResponse {
            timestamp: ts.timestamp(),
            actions: actions.into_iter().map(Into::into).collect(),
        })
    })?)
}
